// GameX - Debug Class Header
//         debugging and general utility/convenience functions and macros
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 


#ifndef GAMEX_DEBUG_H_INCLUDED
#define GAMEX_DEBUG_H_INCLUDED

#include "gamex-global-def.h"

// assert(condition) or ASSERT(condition) will cause a breakpoint if the condition is false, or expand to nothing in Release mode
// verify(condition) or VERIFY(condition) will cause a breakpoint if the condition is false, or expand to (condition) in Release mode
// TRACE(msg) or TRACE(msg, ...) will output something to the Visual Studio output pane only, or expand to nothing in Release mode

// For example:   assert(width > 0);    or    assert(argument != NULL);   or  verify(somethingImportantSucceeds());
// Do this as a sanity check whenever you are assuming anything.
// Calls to assert get removed from release builds, so don't worry about it slowing down your game.
// NOTE: Why bother defining our own assert here? There are 3 very important reasons:
//       1.) We can make the assert usually brings you straight to the relevant line of code in just 1 click instead of the usual 2 clicks and a double-click.
//       2.) We can have different assert behaviour depending on the combination of debug/release build, debugger on/off, and inside/outside library.
//       3.) We can define asserts that output custom messages (assertMsg) or even limited printf-style formatted messages (assertMsg2/3/4)
#undef ASSERT
#undef VERIFY
#undef TRACE
#if defined(_DEBUG) || defined(_LIB)
#	define ASSERT(x) ASSERT_ALWAYS(x)
#	define ASSERT_MSG(x,msg) ASSERT_ALWAYS_MSG(x,msg)
#	define ASSERT_MSG_2(x,msg,arg) ASSERT_ALWAYS_MSG_2(x,msg,arg) // msg must be a string constant
#	define ASSERT_MSG_3(x,msg,arg,arg2) ASSERT_ALWAYS_MSG_3(x,msg,arg,arg2) // msg must be a string constant
#	define ASSERT_MSG_4(x,msg,arg,arg2,arg3) ASSERT_ALWAYS_MSG_3(x,msg,arg,arg2,arg3) // msg must be a string constant
#	define VERIFY(x) ASSERT_ALWAYS(x)
#	define VERIFY_MSG(x,msg) ASSERT_ALWAYS_MSG(x,msg)
#	define VERIFY_MSG_2(x,msg,arg) ASSERT_ALWAYS_MSG_2(x,msg,arg) // msg must be a string constant
#	define VERIFY_MSG_3(x,msg,arg,arg2) ASSERT_ALWAYS_MSG_3(x,msg,arg,arg2) // msg must be a string constant
#	define VERIFY_MSG_4(x,msg,arg,arg2,arg3) ASSERT_ALWAYS_MSG_3(x,msg,arg,arg2,arg3) // msg must be a string constant
#	define TRACE z_trace
	void z_trace(char *messageWithFormat, ...); // used by the TRACE macro
#else
#	define ASSERT(x)
#	define ASSERT_MSG(x,msg)
#	define ASSERT_MSG_2(x,msg,arg)
#	define ASSERT_MSG_3(x,msg,arg,arg2)
#	define ASSERT_MSG_4(x,msg,arg,arg2,arg3)
#	define VERIFY(x) (x)
#	define VERIFY_MSG(x,msg) (x); (msg)
#	define VERIFY_MSG_2(x,msg,arg) (x); (msg); (arg)
#	define VERIFY_MSG_3(x,msg,arg,arg2) (x); (msg); (arg); (arg2)
#	define VERIFY_MSG_4(x,msg,arg,arg2,arg3) (x); (msg); (arg); (arg2); (arg3)
#	define TRACE (true) ? ((void)0) : zz_trace // syntax hack
	inline void zz_trace(char* fmt, ...) {}
#	pragma warning(disable: 4552 4553) // disable "operator has no side effect" warnings
#	pragma warning(disable: 4390) // disable "empty controlled statement found" warnings
#endif
#define ASSERT_BOX_HANDLE_SWITCH_INTERNAL() {case IDABORT: case IDNO: TerminateProcess(GetCurrentProcess(), -1); break; case IDIGNORE: case IDYES: break; default: case IDRETRY: BREAKPOINT_DEBUG(); break;}
#define ASSERT_BOX_AUTO_LINE_HANDLE_INTERNAL() __LINE__, __FILENAME__)) ASSERT_BOX_HANDLE_SWITCH_INTERNAL()
#define ASSERT_DEBUG_AUTOBOX_INTERNAL(x) switch(debug.z_AssertBoxInternal("On line %d of %s,\nthe condition " x "\nfailed to be true.", ASSERT_BOX_AUTO_LINE_HANDLE_INTERNAL()
#define ASSERT_ALWAYS_MSG(x,msg) do{ if(!(x)) IF_DEBUG_CHECK() switch(debug.z_AssertBoxInternal("%s\n\n(line %d of %s)\n", msg, ASSERT_BOX_AUTO_LINE_HANDLE_INTERNAL() } while(false)
#define ASSERT_ALWAYS_MSG_2(x,msg,arg) do{ if(!(x)) IF_DEBUG_CHECK() switch(debug.z_AssertBoxInternal(msg "\n\n(line %d of %s)\n", arg, ASSERT_BOX_AUTO_LINE_HANDLE_INTERNAL() } while(false)
#define ASSERT_ALWAYS_MSG_3(x,msg,arg,arg2) do{ if(!(x)) IF_DEBUG_CHECK() switch(debug.z_AssertBoxInternal(msg "\n\n(line %d of %s)\n", arg, arg2, ASSERT_BOX_AUTO_LINE_HANDLE_INTERNAL() } while(false)
#define ASSERT_ALWAYS_MSG_4(x,msg,arg,arg2,arg3) do{ if(!(x)) IF_DEBUG_CHECK() switch(debug.z_AssertBoxInternal(msg "\n\n(line %d of %s)\n", arg, arg2, arg3, ASSERT_BOX_AUTO_LINE_HANDLE_INTERNAL() } while(false)
#define ASSERT_ALWAYS_AUTO(x) do{ if(!(x)) IF_DEBUG_CHECK() ASSERT_DEBUG_AUTOBOX_INTERNAL( #x ) } while(false)
#if defined(_DEBUG) || !defined(_LIB)
#	define IF_DEBUG_CHECK() if(debug.IsRunningDebug())
#	define ASSERT_ALWAYS(x) do{ if(!(x)) BREAKPOINT_DEBUG() else ASSERT_DEBUG_AUTOBOX_INTERNAL( #x ) } while(false)
#else
#	define IF_DEBUG_CHECK()
#	define ASSERT_ALWAYS(x) do{ if(!(x)) BREAKPOINT_DEBUG(); } while(false)
#endif
#ifdef _LIB // special asserts to be used inside the GameX library that don't get built into the library
#	define assertNoLib(x)
#	define assertMsgNoLib(x,m)
#else
#	define assertNoLib(x) assert(x)
#	define assertMsgNoLib(x,msg) assertMsg(x,msg)
#endif
#define assert ASSERT
#define assertMsg ASSERT_MSG
#define assertMsg2 ASSERT_MSG_2
#define assertMsg3 ASSERT_MSG_3
#define assertMsg4 ASSERT_MSG_4
#define verify VERIFY
#define verifyMsg VERIFY_MSG
#define verifyMsg2 VERIFY_MSG_2
#define verifyMsg3 VERIFY_MSG_3
#define verifyMsg4 VERIFY_MSG_4
#define _assert error DONT_INCLUDE_ASSERT_DOT_H // there's no easy way to prevent assert.h from overriding the assert macro besides targetting something unique to it to make sure it doesn't get included after this
                                                // NOTE: if for some reason you want assert.h to override this custom assert macro, #undef _assert before including it


// BREAKPOINT(): Stops program execution and goes to the debugger.
#ifdef _LIB
#	define BREAKPOINT() BREAKPOINT_DEBUG()
#elif defined(_DEBUG)
#	define BREAKPOINT() BREAKPOINT_ALWAYS()
#else
#	define BREAKPOINT() ((void)0)
#endif
#define BREAKPOINT_DEBUG() if(debug.IsRunningDebug()) CRASH()
#define BREAKPOINT_ALWAYS() BREAKPOINT_DEBUG() else switch(debug.z_AssertBoxInternal("Breakpoint on line %d of %s.", __LINE__, __FILENAME__)) ASSERT_BOX_HANDLE_SWITCH_INTERNAL()
#define CRASH() { zg_EnsureDebuggableEnvironment(); {_asm int 3} }// raises interrupt 3: user breakpoint

extern bool z_GameX_exists;
extern void zg_EnsureDebuggableEnvironment ();
extern int zg_SafeMessageBox (const char* text, const char* title, unsigned int type);

// __FILENAME__ is like the __FILE__ macro, but without the directory information
#define __FILENAME__ REMOVE_TRAILING_SEPARATOR(__FILE__)
#define REMOVE_TRAILING_SEPARATOR(string) (strrchr(string, (int)'\\') ? strrchr(string, (int)'\\')+1 : (strrchr(string, (int)'/') ? strrchr(string, (int)'/')+1 : string))

// make fopen magically work a lot more often than usual
// (recursively searching through or creating directories for it if necessary)
// "filename" is the name of the file, with or without relative or absolute directory stuff
// "mode" is a short string describing the opening method
// commonly used modes: "r" for read text, "rb" for read binary, "w" for erase and write text,
//                      "wb" for erase and write binary, "a" for append text, "ab" for append binary
#include <stdio.h>
#define fopen(filenameString, modeString) debug.z_magic_fopen((filenameString),(modeString))

// version of fopen that doesn't complain when the file isn't found
#define fopen_silent(filename,mode) (((debug.z_silent_fopen = true) == true) ? fopen((filename),(mode)) : NULL)

// NULL-safe fclose
#define fclose(file) do { if(file) { fclose(file); (file) = NULL; } } while(false)

// shows some info about the given variable or expression or type or class
// specifically: what actual type the compiler thinks it is, and its size in bytes
// example use: DISPLAY_TYPE_INFO(2 + 2);  or  DISPLAY_TYPE_INFO(size_t);  or  DISPLAY_TYPE_INFO(GameX);
#define DISPLAY_TYPE_INFO(x) GameX.InfoDialog("\"" #x "\" has type \"%s\" and size %d", TYPENAME_OF(x), sizeof(x))

// shows some info about the given function or method
// example use: DISPLAY_FUNCTION_INFO(GameInit);  or  DISPLAY_FUNCTION_INFO(GameXClass::AccessPixels);
#define DISPLAY_FUNCTION_INFO(x) GameX.InfoDialog("\"" #x "()\" has type \"%s\"", TYPENAME_OF(&(x)))

// returns a string that's the name of the type of the given variable (with no virtual information, unfortunately)
#include "typeinfo.h"
#define TYPENAME_OF(x) typeid(x).name()


// class of the global debug object
class Debugging
{
public:
	// Outputs a message to the debug.log file
	// The message can have printf style formatting.
	// NOTE: Newlines are not automatic; you need to add a "\n" for those.
	void Output (const char* messageWithFormat, ...);

	// output of single specific types for convenience
	inline void Output (int i) { Output("%d ", i); }
	inline void Output (unsigned int i) { Output("%u ", i); }
	inline void Output (float f) { Output("%g ", f); }
	inline void Output (double f) { Output("%lg ", f); }
	inline void Output (bool b) { Output(b ? "true " : "false "); }
	inline void Output (char c) { Output("%d'%c' ", c, c); }
	inline void Output (unsigned char b) { Output("%d'%c' ", b, b); }
	inline void Output (long long i) { Output("%I64d ", i, (double)i); }
	inline void Output (unsigned long long i) { Output("%I64u ", i, (double)i); }
	inline void Output (long double f) { Output("%lg ", f); }
	template <class T> inline void Output (Vec4<T>& v) { Output("(%lg %lg %lg %lg) ", (double)v.x, (double)v.y, (double)v.z, (double)v.w); }
	template <class T> inline void Output (Vec3<T>& v) { Output("(%lg %lg %lg) ", (double)v.x, (double)v.y, (double)v.z); }
	template <class T> inline void Output (Vec2<T>& v) { Output("(%lg %lg) ", (double)v.x, (double)v.y); }
	inline void OutputNewline () { Output("\n"); }

	// Outputs the current time and date and a new line
	void OutputTime (bool andDate = true);

	// Clears/deletes the current debug log file
	void ClearLogFile ();

	// Changes which file to output debug messages to
	void ChangeOutputFile (const char* filename);

	// Flushes output writes to disk
	// and unlocks the debug file so other applications can view it
	void Flush ();

	// Actually opens (double-clicks on) the current debug log file to display it,
	// and pauses the game until it is closed
	void ViewLogFile ();

	// Views or runs any file or application
	void RunAnyFile (const char* filename, bool waitForClose = false);

	// Tells you whether a file exists
	bool DoesFileExist (const char* filename);

	// Executes any command that is valid in cmd.exe
	void ExecuteCmd (const char* command, bool echoAndPauseCmdForUser = false, bool pauseGameUntilCompletion = false);

	// Permanently deletes any file, or erases its contents if it can't be deleted. (be careful!)
	void DeleteAnyFile (const char* filename);

	// Brings up an assertion failure style dialog box, even in release builds
	// If you get here when running in VS.NET: Press F10 to get out, or press Ctrl-Alt-C and use the call stack.
	// (If you called this, consider instead using the global assert() and assertMsg() macros.)
	void AssertBox (const char * messageWithFormat, ...);

	// Tells you whether the game is currently being run from the debugger
	inline bool IsRunningDebug () {return z_debugBuild;}

	// Take a filename and changes it to have the correct path if necessary
	// returns true if the path ends up being valid, false otherwise
	bool FixPath (/* IN & OUT */ char * path);

	Debugging();
	~Debugging();

private:
	char z_logfilename [256];
	FILE* z_file;
	bool z_debugBuild;

public:
	// public so our custom ASSERT_MSG macro can use it
	int z_AssertBoxInternal (const char* messageWithFormat, ...);

	// public so our custom fopen macro can use it
	FILE* z_magic_fopen (char* filename, const char* mode);
	FILE* z_magic_fopen (const char* filename, const char* mode);

	bool z_silent_fopen;
};

#endif